==> fetch.php?media=resources_multi_hardware:lorad_1.0.5.tar.gz: lorad_1.0.5/core_v2/clock.c <==
// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "clock.h"

#include <stdlib.h>
#include <sys/types.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "common/statistics.h"
#include "config.h"

static inline void debug_drift(const char *name, size_t nb_board, struct clock_sync *sync,
			       uint32_t counter, struct timespec *time)
{
	int32_t drift;

	if (sync->sync) {
		drift = (counter - sync->counter) - time_diff_us(&sync->time, time);
		debug("Board #%u: Demodulators synchronized with %s: drift=%ius", nb_board, name, drift);

	} else {
		debug("Board #%u: Demodulators synchronized with %s", nb_board, name);
	}
}

enum sync_status {
	SYNC_UNKNOWN,
	SYNC_DELAYED,
	SYNC_FAILED,
	SYNC_DONE,
};

static enum sync_status synchronize(struct core *core, size_t nb_board)
{
	struct clock_board *clock_board = &core->clock->core->boards[nb_board];
	struct clock_sync monotonic = {0};
	struct clock_sync utc = {0};
	struct clock_sync tai = {0};
	uint32_t dsp_counter;

	// Synchronize TAI + UTC
	if (core->clock->pps_enabled &&
	    sx1301ar_get_trigcnt(nb_board, &tai.counter) == 0) {
		if (sx1301ar_get_trighs(nb_board, &dsp_counter) == 0 &&
		    sx1301ar_synchronize(&clock_board->tref, core->clock->pps_time, tai.counter, dsp_counter, NULL) == 0 &&
		    sx1301ar_set_xtal_err(nb_board, clock_board->tref) == 0) {
			debug("Board #%u: DSP synchronized", nb_board);
		}

		if (clock_get_tai_time(core->clock->tai_offset, &tai.time) == 0) {

			// Synchronize only if we are in the middle of a second
			if (tai.time.tv_nsec < 100000000 ||
			    tai.time.tv_nsec > 900000000) {
				return SYNC_DELAYED;
			}

			tai.time.tv_nsec = 0;
			tai.sync = 1;

			debug_drift("TAI", nb_board, &clock_board->tai, tai.counter, &tai.time);

			// Synchronize UTC from TAI
			if (clock_convert_tai_to_utc(core->clock->tai_offset, &tai.time, &utc.time) == 0) {
				utc.counter = tai.counter;
				utc.sync = 1;

				debug_drift("UTC", nb_board, &clock_board->utc, tai.counter, &utc.time);
			}
		}
	}

	// Synchronize MONOTONIC
	if (sx1301ar_get_instcnt(nb_board, &monotonic.counter) == 0 &&
	    clock_get_monotonic_time(&monotonic.time) == 0) {
		monotonic.sync = 1;

		debug_drift("MONOTONIC", nb_board, &clock_board->monotonic, monotonic.counter, &monotonic.time);
	}

	// Synchronize UTC
	if (!utc.sync &&
	    sx1301ar_get_instcnt(nb_board, &utc.counter) == 0 &&
	    clock_get_utc_time(&utc.time) == 0) {
		utc.sync = 1;

		debug_drift("UTC", nb_board, &clock_board->utc, utc.counter, &utc.time);
	}

	clock_board->tai = tai;
	clock_board->utc = utc;
	clock_board->monotonic = monotonic;

	return (tai.sync || monotonic.sync) ? SYNC_DONE : SYNC_FAILED;
}

static void *sync_thread(void *arg)
{
	struct core *core = arg;
	enum sync_status status;
	size_t synchronized;
	size_t nb_board;

	sleep_sec(5, core->clock->core->running);

	while (core->clock->core->running) {
		for (nb_board = 0, synchronized = 0; core->started && nb_board < core->config->core->nb_boards; nb_board++) {
			for (status = SYNC_UNKNOWN; status != SYNC_DONE;) {
				if (pthread_mutex_lock(&core->clock->core->mutex) == 0) {
					status = synchronize(core, nb_board);
					pthread_mutex_unlock(&core->clock->core->mutex);
				}

				if (status == SYNC_DELAYED) {
					debug("Synchronization delayed");
					sleep_ms(100);

				} else if (status == SYNC_FAILED) {
					warning("Board #%u: Failed to synchronize", nb_board);

				} else if (status == SYNC_DONE) {
					synchronized++;
				}
			}
		}

		if (core->clock->core->synchronized == 0 &&
		    synchronized == nb_board) {
			api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__SYNCHRONIZED);
			core->clock->core->synchronized = 1;
		}

		sleep_sec(15, core->clock->core->running);
	}

	pthread_exit(NULL);
}

struct clock_core *clock_core_init(void)
{
	struct clock_core *clock_core = NULL;
	size_t nb_board;

	if ((clock_core = zalloc(sizeof(*clock_core))) == NULL) {
		eerror("Failed to allocate clock core");
		return NULL;
	}

	if (pthread_mutex_init(&clock_core->mutex, NULL) != 0) {
		error("Failed to initialize clock core mutex");
		free(clock_core);
		return NULL;
	}

	for (nb_board = 0; nb_board < sizeofa(clock_core->boards); nb_board++) {
		clock_core->boards[nb_board].tref = sx1301ar_init_tref();
	}

	return clock_core;
}

int clock_core_start(struct core *core)
{
	size_t nb_board;

	core->clock->core->running = 1;

	if (pthread_create(&core->clock->core->sync_thread, NULL, sync_thread, core) != 0) {
		error("Failed to start clock synchronization thread");
		clock_core_shutdown(core);
		return -1;
	}

	for (nb_board = 0; nb_board < core->config->core->nb_boards; nb_board++) {
		sx1301ar_set_xtal_err(nb_board, core->clock->core->boards[nb_board].tref);
	}

	return 0;
}

int clock_core_join(struct core *core)
{
	if (pthread_join(core->clock->core->sync_thread, NULL) != 0) {
		error("Failed to join clock synchronization thread");
		return -1;
	}

	return 0;
}

void clock_core_shutdown(struct core *core)
{
	if (core && core->clock && core->clock->core) {
		core->clock->core->running = 0;
	}
}

void clock_core_free(struct core *core)
{
	if (core && core->clock) {
		free(core->clock->core);
	}
}

static int get_time_from_counter(struct core *core,
				 uint32_t *orig_counter, struct timespec *orig_tp,
				 uint32_t counter, struct timespec *tp)
{
	uint32_t udiff;

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	udiff = counter - *orig_counter;
	*tp = *orig_tp;
	time_add_us(tp, udiff);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_utc_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!board->utc.sync) {
		return -1;
	}

	return get_time_from_counter(core, &board->utc.counter, &board->utc.time, counter, tp);
}

int clock_get_tai_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!core->clock->pps_enabled ||
	    !board->tai.sync) {
		return -1;
	}

	return get_time_from_counter(core, &board->tai.counter, &board->tai.time, counter, tp);
}

int clock_get_gps_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct timespec tai_time;

	if (clock_get_tai_time_from_counter(core, nb_board, counter, &tai_time) < 0) {
		return -1;
	}

	clock_convert_tai_to_gps(&tai_time, tp);

	return 0;
}

static int get_counter_from_time(struct core *core,
				 struct timespec *orig_tp, uint32_t *orig_counter,
				 struct timespec *tp, uint32_t *counter)
{
	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	*counter = *orig_counter + time_diff_us(orig_tp, tp);

	pthread_mutex_unlock(&core->clock->core->mutex);
	return 0;
}

int clock_get_current_counter(struct core *core, size_t nb_board, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    board->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0 &&
	    get_counter_from_time(core, &board->tai.time, &board->tai.counter, &tp, counter) == 0) {
		return 0;
	}

	if (board->monotonic.sync &&
	    clock_gettime(CLOCK_MONOTONIC, &tp) == 0 &&
	    get_counter_from_time(core, &board->monotonic.time, &board->monotonic.counter, &tp, counter) == 0) {
		return 0;
	}

	return -1;
}

int clock_get_counter_from_utc_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!board->utc.sync) {
		return -1;
	}

	return get_counter_from_time(core, &board->utc.time, &board->utc.counter, tp, counter);
}

int clock_get_counter_from_tai_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!core->clock->pps_enabled ||
	    !board->tai.sync) {
		return -1;
	}

	return get_counter_from_time(core, &board->tai.time, &board->tai.counter, tp, counter);
}

int clock_get_counter_from_gps_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct timespec tai_time;

	clock_convert_gps_to_tai(tp, &tai_time);

	if (clock_get_counter_from_tai_time(core, nb_board, &tai_time, counter) < 0) {
		return -1;
	}

	return 0;
}

int clock_convert_counter(struct core *core, size_t from_board, size_t to_board, uint32_t from_counter, uint32_t *to_counter)
{
	struct clock_board *from = &core->clock->core->boards[from_board];
	struct clock_board *to = &core->clock->core->boards[to_board];
	struct timespec tp;

	if (from_board == to_board) {
		*to_counter = from_counter;
		return 0;
	}

	if (core->clock->pps_enabled &&
	    from->tai.sync && to->tai.sync &&
	    clock_get_tai_time_from_counter(core, from_board, from_counter, &tp) == 0 &&
	    clock_get_counter_from_tai_time(core, to_board, &tp, to_counter) == 0) {
		return 0;
	}

	if (from->monotonic.sync && to->monotonic.sync &&
	    get_time_from_counter(core, &from->monotonic.counter, &from->monotonic.time, from_counter, &tp) == 0 &&
	    get_counter_from_time(core, &to->monotonic.time, &to->monotonic.counter, &tp, to_counter) == 0) {
		return 0;
	}

	return -1;
}

==> fetch.php?media=resources_multi_hardware:lorad_1.2.0.tar.gz: lorad_1.2.0/core_v2/clock.c <==
// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "clock.h"

#include <stdlib.h>
#include <sys/types.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "common/statistics.h"
#include "config.h"

static inline void debug_drift(const char *name, size_t nb_board, struct clock_sync *sync,
			       uint32_t counter, struct timespec *time)
{
	int32_t drift;

	if (sync->sync) {
		drift = (counter - sync->counter) - time_diff_us(&sync->time, time);
		sync->sum_drift += drift;
		debug("Board #%u: Demodulators synchronized with %s: drift=%ius sum=%ius",
		      nb_board, name, drift, sync->sum_drift);

	} else {
		debug("Board #%u: Demodulators synchronized with %s", nb_board, name);
	}
}

enum sync_status {
	SYNC_UNKNOWN,
	SYNC_DELAYED,
	SYNC_FAILED,
	SYNC_DONE,
};

static enum sync_status synchronize(struct core *core, size_t nb_board)
{
	struct clock_board *clock_board = &core->clock->core->boards[nb_board];
	struct clock_sync monotonic = {0};
	struct clock_sync utc = {0};
	struct clock_sync tai = {0};
	uint32_t dsp_counter;

	// Synchronize TAI + UTC
	if (core->clock->pps_enabled &&
	    sx1301ar_get_trigcnt(nb_board, &tai.counter) == 0) {
		if (sx1301ar_get_trighs(nb_board, &dsp_counter) == 0 &&
		    sx1301ar_synchronize(&clock_board->tref, core->clock->pps_time, tai.counter, dsp_counter, NULL) == 0 &&
		    sx1301ar_set_xtal_err(nb_board, clock_board->tref) == 0) {
			debug("Board #%u: DSP synchronized", nb_board);
		}

		if (clock_get_tai_time(core->clock->tai_offset, &tai.time) == 0) {

			// Synchronize only if we are in the middle of a second
			if (tai.time.tv_nsec < 100000000 ||
			    tai.time.tv_nsec > 900000000) {
				return SYNC_DELAYED;
			}

			tai.time.tv_nsec = 0;
			tai.sync = 1;

			debug_drift("TAI", nb_board, &clock_board->tai, tai.counter, &tai.time);

			// Synchronize UTC from TAI
			if (clock_convert_tai_to_utc(core->clock->tai_offset, &tai.time, &utc.time) == 0) {
				utc.counter = tai.counter;
				utc.sync = 1;

				debug_drift("UTC", nb_board, &clock_board->utc, tai.counter, &utc.time);
			}
		}
	}

	// Synchronize MONOTONIC
	if (sx1301ar_get_instcnt(nb_board, &monotonic.counter) == 0 &&
	    clock_get_monotonic_time(&monotonic.time) == 0) {
		monotonic.sync = 1;

		debug_drift("MONOTONIC", nb_board, &clock_board->monotonic, monotonic.counter, &monotonic.time);
	}

	// Synchronize UTC
	if (!utc.sync &&
	    sx1301ar_get_instcnt(nb_board, &utc.counter) == 0 &&
	    clock_get_utc_time(&utc.time) == 0) {
		utc.sync = 1;

		debug_drift("UTC", nb_board, &clock_board->utc, utc.counter, &utc.time);
	}

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return SYNC_FAILED;
	}
	clock_board->tai.sync = tai.sync;
	clock_board->tai.counter = tai.counter;
	clock_board->tai.time = tai.time;
	clock_board->utc.sync = utc.sync;
	clock_board->utc.counter = utc.counter;
	clock_board->utc.time = utc.time;
	clock_board->monotonic.sync = monotonic.sync;
	clock_board->monotonic.counter = monotonic.counter;
	clock_board->monotonic.time = monotonic.time;
	pthread_mutex_unlock(&core->clock->core->mutex);

	return (tai.sync || monotonic.sync) ? SYNC_DONE : SYNC_FAILED;
}

static void *sync_thread(void *arg)
{
	struct core *core = arg;
	enum sync_status status;
	size_t synchronized;
	size_t nb_board;

	sleep_sec(5, core->clock->core->running);

	while (core->clock->core->running) {
		for (nb_board = 0, synchronized = 0; core->started && nb_board < core->config->core->nb_boards; nb_board++) {
			for (status = SYNC_UNKNOWN; status != SYNC_DONE;) {
				status = synchronize(core, nb_board);

				if (status == SYNC_DELAYED) {
					debug("Synchronization delayed");
					sleep_ms(100);

				} else if (status == SYNC_FAILED) {
					warning("Board #%u: Failed to synchronize", nb_board);

				} else if (status == SYNC_DONE) {
					synchronized++;
				}
			}
		}

		if (core->clock->core->synchronized == 0 &&
		    synchronized == nb_board) {
			api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__SYNCHRONIZED);
			core->clock->core->synchronized = 1;
		}

		sleep_sec(15, core->clock->core->running);
	}

	pthread_exit(NULL);
}

struct clock_core *clock_core_init(void)
{
	struct clock_core *clock_core = NULL;
	size_t nb_board;

	if ((clock_core = zalloc(sizeof(*clock_core))) == NULL) {
		eerror("Failed to allocate clock core");
		return NULL;
	}

	if (pthread_mutex_init(&clock_core->mutex, NULL) != 0) {
		error("Failed to initialize clock core mutex");
		free(clock_core);
		return NULL;
	}

	for (nb_board = 0; nb_board < sizeofa(clock_core->boards); nb_board++) {
		clock_core->boards[nb_board].tref = sx1301ar_init_tref();
	}

	return clock_core;
}

int clock_core_start(struct core *core)
{
	size_t nb_board;

	core->clock->core->running = 1;

	if (pthread_create(&core->clock->core->sync_thread, NULL, sync_thread, core) != 0) {
		error("Failed to start clock synchronization thread");
		clock_core_shutdown(core);
		return -1;
	}

	for (nb_board = 0; nb_board < core->config->core->nb_boards; nb_board++) {
		sx1301ar_set_xtal_err(nb_board, core->clock->core->boards[nb_board].tref);
	}

	return 0;
}

int clock_core_join(struct core *core)
{
	if (pthread_join(core->clock->core->sync_thread, NULL) != 0) {
		error("Failed to join clock synchronization thread");
		return -1;
	}

	return 0;
}

void clock_core_shutdown(struct core *core)
{
	if (core && core->clock && core->clock->core) {
		core->clock->core->running = 0;
	}
}

void clock_core_free(struct core *core)
{
	if (core && core->clock) {
		free(core->clock->core);
	}
}

static int get_time_from_counter(struct core *core,
				 uint32_t *orig_counter, struct timespec *orig_tp,
				 uint32_t counter, struct timespec *tp)
{
	uint32_t udiff;

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	udiff = counter - *orig_counter;
	*tp = *orig_tp;
	time_add_us(tp, udiff);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_utc_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!board->utc.sync) {
		return -1;
	}

	return get_time_from_counter(core, &board->utc.counter, &board->utc.time, counter, tp);
}

int clock_get_tai_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!core->clock->pps_enabled ||
	    !board->tai.sync) {
		return -1;
	}

	return get_time_from_counter(core, &board->tai.counter, &board->tai.time, counter, tp);
}

int clock_get_gps_time_from_counter(struct core *core, size_t nb_board, uint32_t counter, struct timespec *tp)
{
	struct timespec tai_time;

	if (clock_get_tai_time_from_counter(core, nb_board, counter, &tai_time) < 0) {
		return -1;
	}

	clock_convert_tai_to_gps(&tai_time, tp);

	return 0;
}

static int get_counter_from_time(struct core *core,
				 struct timespec *orig_tp, uint32_t *orig_counter,
				 struct timespec *tp, uint32_t *counter)
{
	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	*counter = *orig_counter + time_diff_us(orig_tp, tp);

	pthread_mutex_unlock(&core->clock->core->mutex);
	return 0;
}

int clock_get_current_counter(struct core *core, size_t nb_board, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    board->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0 &&
	    get_counter_from_time(core, &board->tai.time, &board->tai.counter, &tp, counter) == 0) {
		return 0;
	}

	if (board->monotonic.sync &&
	    clock_gettime(CLOCK_MONOTONIC, &tp) == 0 &&
	    get_counter_from_time(core, &board->monotonic.time, &board->monotonic.counter, &tp, counter) == 0) {
		return 0;
	}

	return -1;
}

int clock_get_counter_from_utc_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!board->utc.sync) {
		return -1;
	}

	return get_counter_from_time(core, &board->utc.time, &board->utc.counter, tp, counter);
}

int clock_get_counter_from_tai_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct clock_board *board = &core->clock->core->boards[nb_board];

	if (!core->clock->pps_enabled ||
	    !board->tai.sync) {
		return -1;
	}

	return get_counter_from_time(core, &board->tai.time, &board->tai.counter, tp, counter);
}

int clock_get_counter_from_gps_time(struct core *core, size_t nb_board, struct timespec *tp, uint32_t *counter)
{
	struct timespec tai_time;

	clock_convert_gps_to_tai(tp, &tai_time);

	if (clock_get_counter_from_tai_time(core, nb_board, &tai_time, counter) < 0) {
		return -1;
	}

	return 0;
}

int clock_convert_counter(struct core *core, size_t from_board, size_t to_board, uint32_t from_counter, uint32_t *to_counter)
{
	struct clock_board *from = &core->clock->core->boards[from_board];
	struct clock_board *to = &core->clock->core->boards[to_board];
	struct timespec tp;

	if (from_board == to_board) {
		*to_counter = from_counter;
		return 0;
	}

	if (core->clock->pps_enabled &&
	    from->tai.sync && to->tai.sync &&
	    clock_get_tai_time_from_counter(core, from_board, from_counter, &tp) == 0 &&
	    clock_get_counter_from_tai_time(core, to_board, &tp, to_counter) == 0) {
		return 0;
	}

	if (from->monotonic.sync && to->monotonic.sync &&
	    get_time_from_counter(core, &from->monotonic.counter, &from->monotonic.time, from_counter, &tp) == 0 &&
	    get_counter_from_time(core, &to->monotonic.time, &to->monotonic.counter, &tp, to_counter) == 0) {
		return 0;
	}

	return -1;
}
